# This file is a part of alpha.

import ambient
import os.path

# Singleton design pattern implementation from "Python Cookbook 2nd. Ed. 6.15"
class Singleton(object):
	def __new__(klass, *args, **kwargs):
		if '_inst' not in vars(klass):
			klass._inst = super(Singleton, klass).__new__(klass, *args, **kwargs)
		return klass._inst

class MRUManager(Singleton):
	def __init__(self, maximum_number = 4):
		self.__filenames = list()
		self.__maximum_number = maximum_number
		self.__menu = ambient.ui.PopupMenu(self.__update_menu)
		self.__changed = True
		try:
			self.read_file()
		except IOError:
			pass
	def __del__(self):
		try:
			self.write_file()
		except IOError:
			ambient.error('Unabled to save the most recent used file list.')
	def __getitem__(self, i):
		return self.__filenames[i]
	def __len__(self):
		return len(self.__filenames)
	def get_menu(self):
		return self.__menu
	def insert(self, filename):
		normalized = os.path.normpath(os.path.normcase(filename))
		for i, s in enumerate(self.__filenames):
			if normalized == os.path.normpath(os.path.normcase(s)):
				del self.__filenames[i]
				self.__filenames.insert(0, filename)
				return
		self.__filenames.insert(0, filename)
		if len(self.__filenames) > self.__maximum_number:
			del self.__filenames[-1]
		self.__changed = True
	def read_file(self):
		f = open('mru', 'r')
		i = 0
		filenames = list()
		for line in f:
			if i > self.__maximum_number:
				break
			filenames.append(line)
			i += 1
		self.__filenames = filenames
		self.__changed = True
		f.close()
	def write_file(self):
		f = open('mru', 'w')
		f.writelines(self.__changed)
		f.close()
	def set_maximum_number(self, n):
		if n < 1:
			raise ValueError(n)
		if len(self.__filenames) > n:
			del self.__filename[n:]
			self.__changed = True
	def __update_menu(self, unused, m):
		if not self.__changed:
			return
		m.clear()
		if len(self.__filenames) != 0:
			for filename in self.__filenames:
				m.append(0, filename, lambda : open_file(filename))
		else:
			m.append(0, '(None)', None)
			m.enable(0, False)
		self.__changed = False

def new_file():
	'''Creates a new document.'''
	ambient.active_window().select(ambient.buffers().add_new())

def new_file_dialog():
	'''Creates a new document with specified encoding, newline and mode.'''
	new_buffer = ambient.buffers().add_new_dialog()
	if new_buffer is not None:
		ambient.active_window().select(new_buffer)

def handle_file_io_error(error, filename):
	ambient.error(filename + '\n\n' + unicode(error))

def _open_file(filename, encoding, encoding_substitution_policy, lock_type, lock_only_as_editing, as_read_only):
	if os.path.exists(filename):
		filename = os.path.normcase(os.path.realpath(filename))
	if os.path.isdir(filename):
		ambient.error(filename + '\n\n This is a directory.')
	buffer = ambient.buffers().for_filename(filename)
	if buffer is not None:
		if not buffer.verify_time_stamp():
			return buffer
		ambient.active_window().select(buffer)
	else:
		last_buffer = ambient.active_buffer()
		buffer = ambient.buffers().add_new()
		ambient.active_window().select(buffer)
		try:
			buffer.bind_file(filename)
			buffer.revert_to_file(encoding, encoding_substitution_policy)
#			buffer.lock_file(lock_mode, )
		except:
			ambient.active_window().select(last_buffer)
			buffer.close()
			raise
	if as_read_only:
		buffer.read_only = True
	return buffer

def open_file(filename, encoding = 'UniversalAutoDetect', lock_type = ambient.FileLockMode.no_lock, lock_only_as_editing = False, as_read_only = False):
	'''Opens an existing file, and activates and returns the opened buffer.'''
	from ambient.ui import MessageBox
	buffer = None
	while True:
		try:
			buffer = _open_file(filename, encoding, ambient.EncodingSubstitutionPolicy.dont_substitute, lock_type, lock_only_as_editing, as_read_only)
		except ambient.UnmappableCharacterError:
			modified_encoding = encoding
			user_answer = MessageBox.show('%s\nEncoding: %s\n\nThe file contains unconvertable characters under specified encoding.\nSaving may destroy contents of the file.\nDo you want to change encoding?' % (filename, encoding), buttons = MessageBox.Buttons.yes_no_cancel, icon = MessageBox.Icon.exclamation)
		except ambient.MalformedInputError:
			modified_encoding = encoding
			user_answer = MessageBox.show('%s\nEncoding: %s\n\nThe file contains malformed byte sequence as the specified encoding. Try the other encoding.' % (filename, encoding), buttons = MessageBox.Buttons.ok_cancel, icon = MessageBox.Icon.exclamation)
		except IOError(e):
			handle_file_io_error(e, filename)
			return None
		if buffer is None:
			if user_answer == MessageBox.Result.yes or user_answer == MessageBox.Result.ok:
				# the user wants to change the encoding
				encoding = ambient.ui.encodings_dialog(encoding, True)
				if len(encoding) == 0:
					return None
				continue
			elif user_answer == MessageBox.Result.no:
				try:
					buffer = _open_file(filename, encoding, ambient.EncodingSubstitutionPolicy.replace_unmappable_characters, lock_type, lock_only_as_editing, as_read_only)
				except ambient.MalformedInputError:
					MessageBox.show('%s\nEncoding: %s\n\nThe file contains malformed byte sequence as the specified encoding. Try the other encoding.' % (filename, encoding), buttons = MessageBox.Buttons.ok, icon = MessageBox.Icon.exclamation)
					return None
			else:
				return None
		break
	ambient.active_window().select(buffer)
	return buffer

def open_file_dialog():
	'''Opens an existing file(s).'''
	import ambient.ui
	files = ambient.ui.open_file_dialog()
	if len(files) == 0:
		return
	for file in files:
		if open_file(file['filename'], file['encoding'], ambient.FileLockMode.no_lock, False, file['read_only']) is None:
			return
		MRUManager().insert(ambient.active_buffer().location)

def _reopen_file(buffer = None, change_encoding = False):
	from ambient.ui import MessageBox
	if buffer is None:
		buffer = ambient.active_buffer()
	if not buffer.is_bound_to_file():
		return False
	if buffer.is_modified() and MessageBox.show(_('The file is modified.\nIf you reopen, the modification will be lost. Do you want to do?'), buttons = MessageBox.Buttons.yes_no, icon = MessageBox.Icon.question) == MessageBox.Result.no:
		return False	# the user canceled

	encoding = buffer.encoding
	if change_encoding:
		encoding = ambient.ui.encodings_dialog(encoding, True)
		if len(encoding) == 0:
			return False
	# todo: continue to implement...

def reopen_file():
	'''Reopens the document.'''
	return _reopen_file()

def reopen_with_different_encoding():
	'''Reopens the document with different encoding.'''
	return _reopen_file(change_encoding = True)

def close_buffer(buffer = None):
	'''Closes the document.'''
	if buffer is None:
		buffer = ambient.active_buffer()
	if buffer.is_modified():
		ambient.active_window().select(buffer)
		buffer_name = buffer.location
		if len(buffer_name) == 0:
		    buffer_name = buffer.name
		from ambient.ui import MessageBox
		answer = MessageBox.show('%s\n\nThis buffer is modified. Do you want to save?' % buffer_name, buttons = MessageBox.Buttons.yes_no_cancel, icon = MessageBox.Icon.exclamation)
		if answer == MessageBox.Result.cancel:
			return False
		elif answer == MessageBox.Result.yes:
			if not save_buffer(buffer):
				return False
	buffer.close()
	return True

def _close_all_buffers(except_active = False):
	active_was_front = ambient.active_buffer() is ambient.buffers()[0]

	# first, erase all buffers no need to save
	i = len(ambient.buffers()) - 1
	while i >= 0:
		buffer = ambient.buffers()[i]
		if not buffer.is_modified() and (not except_active or buffer is ambient.active_buffer()):
			buffer.close()
		i -= 1

	# exit if there is no buffer unmodified
	if len(ambient.buffers()) == 1 and (except_active or not ambient.buffers()[0].is_modified()):
		return True

	# there is one modified buffer => normal confirm dialog box and exit
	if (except_active and len(ambient.buffers()) == 2) or (not except_active and len(ambient.buffers()) == 1):
		if not except_active:
			buffer = ambient.buffers()[0]
		elif active_was_front:
			buffer = ambient.buffers()[1]
		else:
			buffer = ambient.buffers()[0]
		if buffer.is_modified():
			if except_active:
				ambient.active_window().select(buffer)
			return close_buffer()

	# save several buffers...
	if not ambient.buffers().save_some_dialog():
		return False
	ambient.buffers().close_all()
	return True

def close_all_buffers():
	'''Closes all documents.'''
	return _close_all_buffers(False)

def close_all_inactive_buffers():
	'''Closes all inactive documents.'''
	return _close_all_buffers(True)

def _try_save_buffer(buffer, encoding, newlines, encoding_substitution_policy, write_unicode_byte_order_mark):
	while True:
		try:
			buffer.save(encoding, newlines, encoding_substitution_policy, write_unicode_byte_order_mark)
		except ambient.UnmappableCharacterError:
			from ambient.ui import MessageBox
			user_answer = MessageBox.show('%s\nEncoding: %s\n\nThe buffer contains unconvertable characters under specified encoding.\nThese characters will be written incorrectly.\nDo you want to change encoding?' % (buffer.location, encoding), buttons = MessageBox.Buttons.yes_no_cancel, icon = MessageBox.Icon.exclamation)
			if user_answer == MessageBox.Result.yes:
				# the user wants to change the encoding
				encoding = ambient.ui.encodings_dialog(encoding, False)
				if len(encoding) == 0:
					return False
				continue
			elif user_answer == MessageBox.Result.no:
				encoding_substitution_policy = ambient.EncodingSubstitutionPolicy.replace_unmappable_characters
				buffer.save(encoding, newlines, encoding_substitution_policy, write_unicode_byte_order_mark)
				break
			else:
				return False
		except IOError(e):
			handle_io_error(e)
			return False
		break
	MRUManager().insert(buffer.location)
	return True

def save_buffer(buffer = None, encoding = '', newlines = ambient.Newline.raw_value, encoding_substitution_policy = ambient.EncodingSubstitutionPolicy.dont_substitute, write_unicode_byte_order_mark = False):
	'''Writes the content of the buffer into the bound file.'''
	if buffer is None:
		buffer = ambient.active_buffer()
	if not buffer.is_modified():
		return True
	elif (not buffer.is_bound_to_file()) or (not os.path.exists(buffer.location)):
		return save_buffer_dialog(buffer)
	return _try_save_buffer(buffer, encoding, newlines, encoding_substitution_policy, write_unicode_byte_order_mark)

def save_buffer_dialog(buffer = None, encoding_substitution_policy = ambient.EncodingSubstitutionPolicy.dont_substitute, write_unicode_byte_order_mark = False):
	if buffer is None:
		buffer = ambient.active_buffer()
	result = ambient.ui.save_file_dialog('', buffer.encoding, buffer.newline)
	if result is not None:
		old_filename = buffer.location
		modified = buffer.is_modified()
		buffer.bind_file(result['filename'])
		if _try_save_buffer(buffer, result['encoding'], result['newline'], encoding_substitution_policy, write_unicode_byte_order_mark):
			return True
		else:
			buffer.bind_file(old_filename)
			if not modified:
				buffer.mark_unmodified()
	return False

def save_some_buffers(verbose = True, unbound_buffer_handler = _try_save_buffer):
	from ambient.ui import MessageBox
	saved_buffers = 0
	for buffer in ambient.buffers():
		if buffer.is_modified():
			ambient.active_window().select(buffer)
			if not buffer.is_bound_to_file() or verbose:
				user_answer = MessageBox.show('%s\n\nDo you want to save?' % buffer.name, buttons = MessageBox.Buttons.yes_no_cancel, icon = MessageBox.Icon.question)
				if user_answer == MessageBox.Result.no:
					continue
				elif user_answer == MessageBox.Result.cancel:
					break
			if not buffer.is_bound_to_file():
				succeeded = unbound_buffer_handler(buffer)
			else:
				succeeded = _try_save_buffer(buffer)
			if not succeeded:
				break
			saved_buffers += 1
	if saved_buffers > 0:
		MessageBox.show('Saved %d buffers.' % saved_buffers)

def save_all_buffers():
	return save_some_buffers(False, save_buffer)

def print_buffer_dialog():
	import ambient.printing
	return ambient.printing.print_buffer(show_dialog = True)


from ambient.bindings import *
k = InputMappingScheme.get()
k.define(KeyStroke('N', ModifierKey.ctrl), new_file)
k.define(KeyStroke('O', ModifierKey.ctrl), open_file_dialog)
k.define(KeyStroke(NaturalKey.f4, ModifierKey.ctrl), close_buffer)
k.define(KeyStroke('S', ModifierKey.ctrl), save_buffer)
k.define(KeyStroke('S', ModifierKey.ctrl_shift), save_all_buffers)
k.define(KeyStroke('P', ModifierKey.ctrl), print_buffer_dialog)
